package com.recordingblogs.squarewave;

import com.recordingblogs.orgenerator.*;

public class SquareWave implements GeneratorInterface
{
	/**
	 * The minimum allowed frequency of the generated square wave, in Hz
	 */
	public static final float MINFREQUENCY = 20;
	
	/**
	 * The maximum allowed frequency of the generated square wave, in Hz
	 */
	public static final float MAXFREQUENCY = 22000;

	/**
	 * The sampling rate of the generated square wave.  This will be 
	 * provided by Orinj with a call to the function setAudioFormat below
	 */
	private float m_samplingRate;
	
	/**
	 * The sampling resolution of the generated square wave.  This will be 
	 * provided by Orinj with a call to the function setAudioFormat below
	 */
	private int m_bitResolution;
	
	/**
	 * The number of audio channels in the generated square wave.  This will be 
	 * provided by Orinj with a call to the function setAudioFormat below
	 */
	private int m_channels;

	/**
	 * The frequency of the generated square wave.  This will be set by
	 * the graphical user interface SquareWavePanel
	 */
	private float m_frequency;
	
	/**
	 * The current time of the buffers sent to getData below
	 */
	private double m_time;

	/**
	 * The default and only contstructor
	 */
	public SquareWave()
	{
		m_samplingRate = 44100F;
		m_bitResolution = 16;
		m_channels = 2;
		m_frequency = 440F;
		m_time = -1;
	}

	/**
	 * Get the frequency of the square wave to be generated.  This is used by
	 * the graphical user interface SquareWavePanel
	 * @return The frequency of the square wave to be generated
	 */
	public float getFrequency()
	{
		return m_frequency;
	}

	/**
	 * Set the frequency of the square wave to be generated.  The frequency
	 * will be set with a call to this function by the graphical user
	 * interface SquareWavePanel
	 * @param frequency The new frequency
	 */
	public void setFrequency(float frequency)
	{
    	if (frequency < MINFREQUENCY || frequency > MAXFREQUENCY)
    		System.out.println("Error in SquareWave::setFrequency: Error in frequency");
    	frequency = Math.min(MAXFREQUENCY, Math.max(MINFREQUENCY, frequency));
		m_frequency = frequency;
	}

	/**
	 * Set the audio format of the square wave to be generated.  This function
	 * will be called by Orinj
	 * @param samplingRate The new sampling rate
	 * @param bitResolution The new sampling resolution
	 * @param channels The new number of channels
	 */
	public void setAudioFormat(float samplingRate, int bitResolution, int channels)
	{
		m_samplingRate = samplingRate;
		m_bitResolution = bitResolution;
		m_channels = channels;
	}

	/**
	 * Set the language of the graphical user interface. Implementing languages
	 * is not required and languages are not implemented in this example
	 * @param languageCode The three-letter ISO 639-2 code of the new language
	 */
    public void setLanguage(String languageCode)
    {
    	// nothing to do; not implemented in this example
    }

    /**
     * Get a buffer of audio data to be written to the generated wave.  
     * @param buf The buffer, in which generated audio data should be placed
     * @param gain The gain (a number between 0 and 1) that should be applied to the
     *             generated audio data, before these data are placed in buffer
     * @return The number of audio bytes generated
     */
	public int getData(byte [] buf, double gain)
	{
		// block align is the number of bytes required for one sample 
		// if there are 2 channels and the sampling resolution is 16 bits, then
		// block align is 4 = 2 * 16 / 8.  That means that each sample will be
		// written in four bytes (2 for the left channela and 2 for the right
		// channel)
		int blockAlign = m_channels * m_bitResolution / 8;
		
		// if m_time is less than 0, then this is the first call to getData and 
		// the time should be initialized to zero.  Otherwise, the time should
		// be incremented by the time length of the buffer
		// You should expect that only one wave will be generated by an 
		// instantiation of a SquareWave object and so this below is safe
		if (m_time < 0)
			m_time = 0;
		else
			m_time += (double) buf.length / ((float) blockAlign * m_samplingRate);

		// populate the audio buffer with data
		for(int i = 0; i < buf.length; i += blockAlign)
		{
			// generate a sine wave with the desired frequency (and phase, given m_time)
			double value = Math.sin(2 * Math.PI * (m_time + (((float) i / blockAlign)/ m_samplingRate)) * m_frequency);
			
			// take the sign of the sine wave to create a square wave
			if (value > 0)
				value = 1;
			else if (value < 0)
				value = -1;

			// convert the result to audio bytes appropriate for the wave file 
			// format, depending on the sampling resolution
            if (m_bitResolution == 8)
            {
                value = Math.min(Math.max(value * gain * Byte.MAX_VALUE, Byte.MIN_VALUE), Byte.MAX_VALUE);
	            buf[i] = (byte) (((int) value + 128) & 0xff);
	            if (m_channels > 1)
	            	buf[i + 1] = buf[i];
            }
            else if (m_bitResolution == 16)
            {
	            value = Math.max(Math.min(value * gain * Short.MAX_VALUE, Short.MAX_VALUE), Short.MIN_VALUE);
	        	buf[i] = (byte)((int) value & 0xff);
	        	buf[i + 1] = (byte)((int) value >>> 8 & 0xff);
	        	if (m_channels > 1)
	        	{
	        		buf[i + 2] = buf[i];
	        		buf[i + 3] = buf[i + 1];
	        	}
            }
            else if (m_bitResolution == 24)
            {
                value = Math.min(Math.max(value * gain * 8388607F, -8388607F), 8388607F);
	            buf[i] = (byte)((int) value & 0xff);
	            buf[i + 1] = (byte)((int) value >>> 8 & 0xff);
	            buf[i + 2] = (byte)((int) value >>> 16 & 0xff);
	        	if (m_channels > 1)
	        	{
	        		buf[i + 3] = buf[i];
	        		buf[i + 4] = buf[i + 1];
	        		buf[i + 5] = buf[i + 2];
	        	}
            }
            else
            {
            	// in Orinj, 32-bit data is always Microsoft IEEE float, not Microsoft PCM
                value = Math.min(Math.max(value * gain, -1F), 1.0F);
	            int iTo = Float.floatToIntBits((float) value);
	            buf[i] = (byte)(iTo & 0xff);
	            buf[i + 1] = (byte)(iTo >>> 8 & 0xff);
	            buf[i + 2] = (byte)(iTo >>> 16 & 0xff);
	            buf[i + 3] = (byte)(iTo >>> 24 & 0xff);
	        	if (m_channels > 1)
	        	{
	        		buf[i + 4] = buf[i];
	        		buf[i + 5] = buf[i + 1];
	        		buf[i + 6] = buf[i + 2];
	        		buf[i + 7] = buf[i + 3];
	        	}
            }
		}

		// return the number of bytes generated.  In this example, this is 
		// always the length of the audio buffer
		return buf.length;
	}
}
